#!/usr/bin/env python
# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import logging
import string
import sys
import os
import time
import platform

import usb.core
import usb.util
import usb.backend.libusb1
import serial.tools.list_ports

# fix of broken paths for windows
if not sys.path:
    sys.path = []
path = os.getcwd()
if path not in sys.path:
    sys.path.append(path)

import config


class BaseDetector:

    @staticmethod
    def format_vid_or_pid(vid_or_pid): #cuts "0x", fill with zeroes if needed, doing case up
        if vid_or_pid:
            return hex(vid_or_pid)[2:].zfill(4).upper()

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.backward_compatible_serials = config.get_settings().get("backward_compatible_serials")

    # find printers with same VID, PID, SNR and include COM or PRT name in their SNRs
    def modify_identical_printers_id(self, printers, modification_attr):
        for checked_printer in printers:
            for printer in printers:
                if printer['VID'] == checked_printer['VID'] and\
                   printer['PID'] == checked_printer['PID'] and\
                   printer['SNR'] == checked_printer['SNR'] and\
                   printer[modification_attr] != checked_printer[modification_attr]:
                      self.update_duplicate_printer_info(printer, modification_attr)
                      self.update_duplicate_printer_info(checked_printer, modification_attr)

    def update_duplicate_printer_info(self, printer, modification_attr):
        printer['ID_CLONE_WARNING'] = True
        if printer[modification_attr]:
            if not printer['SNR']:
                printer['SNR'] = ''
            printer['SNR'] = printer['SNR'] + printer[modification_attr]


class PySerialDetector(BaseDetector):

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.info("PySerial version:" + serial.__version__)
        super().__init__()

    def get_printers_list(self, ignore_profiles=False, add_descriptions=False):
        printers = []
        try:
            serial_ports = serial.tools.list_ports.comports()
        except Exception as e:
            self.logger.warning(e)
        for port in serial_ports:
            if port.vid and port.pid:
                usb_info = {
                       'VID': self.format_vid_or_pid(port.vid),\
                       'PID': self.format_vid_or_pid(port.pid),\
                       'SNR': port.serial_number,\
                       'PRT': port.location.split(":")[0], # remove endings like ":1.0"\
                       'COM': port.device}
                if not usb_info['SNR']:
                    usb_info['SNR'] = ''
                if add_descriptions:
                    manufacturer = port.manufacturer
                    product = port.product
                    if not manufacturer:
                        manufacturer = ''
                    if not product:
                        product = ''
                    usb_info.update({"manufacturer": manufacturer, "product": product})
                printers.append(usb_info)
        if self.backward_compatible_serials or not ['PRT']:
            self.modify_identical_printers_id(printers, 'COM')
        else:
            self.modify_identical_printers_id(printers, 'PRT')
        return printers

class PyUSBDetector(BaseDetector):
    #NOTE Don't use try:... except (usb.core.USBError) for pyusb calls, because pyusb is unstable(up to v1.0.2)
    #and can produce all kinds of unpredictable exceptions

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.info("PyUSB version:" + usb.__version__)
        self.libusb_backend = usb.backend.libusb1.get_backend(find_library=self.get_libusb_path)
        super().__init__()

    def get_libusb_path(self, lib):
        if sys.platform.startswith('win'):
            python_version = platform.architecture()[0]
            if '64' in python_version:
                libusb_name = 'libusb-1.0-64.dll'
            else:
                libusb_name = 'libusb-1.0.dll'
            pythons_folder = os.path.dirname(os.path.abspath(sys.executable))
            embeded_python_backend_path = os.path.join(pythons_folder, libusb_name)
            if os.path.isfile(embeded_python_backend_path):
                return embeded_python_backend_path
            local_backend_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), libusb_name)
            if os.path.isfile(local_backend_path):
                return local_backend_path

    def detect_devices(self):
        retries = 5
        while retries: # the retries is for mac, because it can suddenly refuse to detect usb devices
            try:
                devices = usb.core.find(find_all=True, backend=self.libusb_backend)
            except Exception as e:
                self.logger.exception("Exception in pyusb: %s" % e)
                devices = None
            if not devices:
                if retries == 5:
                    message = "Warning: no USB devices detected. This is an error. Detection retries left: %d" % retries
                    self.logger.warning(message)
                time.sleep(0.5)
                retries -= 1
            else:
                return devices
        else:
            return []

    def get_printers_list(self, ignore_profiles=False, add_descriptions=False):
        printers = []
        for dev in self.detect_devices():
            try:
                VID = self.format_vid_or_pid(dev.idVendor)
                PID = self.format_vid_or_pid(dev.idProduct)
            except:
                continue #anything without VID and PID is useless for us
            try:
                bus = dev.bus
                port_numbers = dev.port_numbers
                if port_numbers is None:
                    port_numbers = (dev.port_number,)
                PRT = "%d-%d" % (bus, port_numbers[0])
                PRT += ".%d" * len(port_numbers[1:]) % port_numbers[1:]
            except:  
                PRT = ""
            try:
                SNR = usb.util.get_string(dev, dev.iSerialNumber)
                for symbol in SNR:
                    if symbol not in string.printable:
                        raise RuntimeError()
            except:
                SNR = ""
            usb_info = {'VID': VID, 'PID': PID, 'SNR': SNR, 'PRT': PRT}
            if add_descriptions:
                try:
                    usb_info["manufacturer"] = str(dev.manufacturer)
                except: #some device descriptors will make libusb to crash on this action
                    usb_info["manufacturer"] = ''
                try:
                    usb_info["product"] = str(dev.product)
                except: #some device descriptors will make libusb to crash on this action
                    usb_info["product"] = ''
            printers.append(usb_info)
        self.modify_identical_printers_id(printers, 'PRT')
        return printers


class USBDetector:
    
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.forced_types = config.get_settings().get('forced_types', {})
        self.no_empty_serials = config.get_settings().get('no_empty_serials')
        self.detectors = []
        self.detectors.append(PySerialDetector()) #PySerial should be first 
        self.detectors.append(PyUSBDetector())

    def add_forced_type(self, printer):
        forced_type = self.forced_types.get(printer['VID'] + ":" + printer['PID'])
        if forced_type:
            printer['forced_type'] = forced_type

    def get_printers_list(self, ignore_profiles=False, add_descriptions=False):
        """Collects together devices from all of detector
           Filters printers already detected by previous detectors"""
        printers = []
        profiles = config.get_profiles()
        for detector in self.detectors:
            detected_printers = detector.get_printers_list(ignore_profiles, add_descriptions)
            for printer in detected_printers:
                already_exists = False
                got_profile = False
                for existing_printer in list(printers): # list creates different copy of printers
                    if printer['VID'] == existing_printer['VID'] and\
                       printer['PID'] == existing_printer['PID']:
                           already_exists = True
                if already_exists:
                    continue
                for profile in profiles:
                    if [printer['VID'], printer['PID']] in profile['vids_pids']:
                        if profile.get('crazy_serial_number') and not printer.get('ID_CLONE_WARNING'):
                            printer['SNR'] = ''
                        got_profile = True
                        # break # not break here to make crazy_serial_number check more reliable
                if not printer['SNR']:
                    if self.no_empty_serials:
                        printer['SNR'] = printer['PRT']
                if not (got_profile or ignore_profiles):
                    continue
                self.add_forced_type(printer)
                printers.append(printer)
        return printers

if __name__ == '__main__':
    print('PyUSB v' + usb.__version__)
    print('PySerial v' + serial.__version__)
    detector = USBDetector()
    printers = detector.get_printers_list(ignore_profiles=True, add_descriptions=True)
    print("\nAll devices:")
    if printers:
        for printer in printers:
            print(printer)
    else:
        print(None)
    if config.get_profiles(): #we dont need this we we cant import config and profiles
        printers = detector.get_printers_list()
        print("\nPrinters with profiles:")
        if printers:
            for printer in printers:
                print(printer)
        else:
            print(None)
    printers = [x for x in printers if x.get('COM')]
    print("\nDevices with serial port:")
    if printers:
        for printer in printers:
            print(printer)
    else:
        print(None)
    print("\nCOM port devices by pyserial:")
    for port in serial.tools.list_ports.comports():
        print("Device:%s HWID:%s" % (port.device, port.hwid))
